"""Abstract base class for preprocessing methods."""
from abc import ABC, abstractmethod
from typing import Annotated, Any

from ML_management.mlmanagement.jsonschema_inference import SkipJsonSchema
from ML_management.model.patterns.model_pattern import Model

Tensor = Any


class Preprocessor(Model, ABC):
    """Abstract class for model that performs preprocessing."""

    @abstractmethod
    def preprocess(self, input_batch: Annotated[Tensor, SkipJsonSchema], **kwargs) -> Tensor:
        """Perform data preprocessing."""
        raise NotImplementedError

    def predict_function(self, input_batch: Annotated[Tensor, SkipJsonSchema]):
        return self.preprocess(input_batch)



from typing import List, Optional

import pandas as pd
from sgqlc.operation import Operation

from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import DatasetLoaderMethodName
from ML_management.graphql import schema
from ML_management.graphql.schema import (
    DatasetLoaderInfo,
    DatasetLoaderVersionInfo,
    ObjectFilterSettings,
    ObjectVersionFilterSettings,
    TagFilterSettings,
    TimestampInterval,
    UpdateObjectForm,
)
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.sdk.sdk import _print_params_by_schema, _to_datetime


def list_dataset_loader() -> pd.DataFrame:
    """
    List available dataset_loaders.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with list of available dataset_loaders.
    """
    op = Operation(schema.Query)
    op.list_dataset_loader.name()
    op.list_dataset_loader.description()
    op.list_dataset_loader.creation_timestamp()
    op.list_dataset_loader.last_updated_timestamp()
    json_data = send_graphql_request(op)
    df = pd.DataFrame.from_dict(json_data["listDatasetLoader"])
    if not df.empty:
        df = _to_datetime(df, ["creationTimestamp", "lastUpdatedTimestamp"])
    return df


def set_dataset_loader_tags(name: str, key: str, values: list[str]) -> DatasetLoaderInfo:
    """
    Set dataset loader tags.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    values: list[str]
        Value tag.

    Returns
    -------
    DatasetLoaderInfo
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_tag = op.set_dataset_loader_tags(name=name, key=key, values=values)
    set_tag.name()
    set_tag.tags()
    dataset_loader = send_graphql_request(op=op, json_response=False).set_dataset_loader_tags
    return dataset_loader


def reset_dataset_loader_tags(
    name: str, key: str, values: list[str], new_key: Optional[str] = None
) -> DatasetLoaderInfo:
    """
    Reset dataset loader tags.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.


    Returns
    -------
    DatasetLoaderInfo
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)
    set_tag = op.reset_dataset_loader_tags(name=name, key=key, values=values, new_key=new_key)
    set_tag.name()
    set_tag.tags()
    dataset_loader = send_graphql_request(op=op, json_response=False).reset_dataset_loader_tags
    return dataset_loader


def delete_dataset_loader_tag(name: str, key: str, value: Optional[str] = None) -> DatasetLoaderInfo:
    """
    Delete dataset loader tag.

    Parameters
    ----------
    name: str
        Name of the model.
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.
    Returns
    -------
    DatasetLoaderInfo
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)
    delete_tag = op.delete_dataset_loader_tag(name=name, key=key, value=value)
    delete_tag.name()
    delete_tag.tags()
    dataset_loader = send_graphql_request(op=op, json_response=False).delete_dataset_loader_tag
    return dataset_loader


def set_dataset_loader_description(name: str, description: str) -> DatasetLoaderInfo:
    """
    Set dataset loader description.

    Parameters
    ----------
    name: str
        Name of the model.
    description: str
        Description model.

    Returns
    -------
    DatasetLoaderInfo
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)

    set_description = op.update_dataset_loader(
        name=name,
        update_dataset_loader_form=UpdateObjectForm(new_description=description),
    )
    set_description.name()
    set_description.description()
    dataset_loader = send_graphql_request(op=op, json_response=False).update_dataset_loader
    return dataset_loader


def set_dataset_loader_visibility(name: str, visibility: VisibilityOptions) -> DatasetLoaderInfo:
    """
    Set dataset loader visibility.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    visibility: VisibilityOptions
        Visibility of the dataset loader.

    Returns
    -------
    DatasetLoaderInfo
        DatasetLoader instance with meta information.
    """
    op = Operation(schema.Mutation)

    set_visibility = op.update_dataset_loader(
        name=name,
        update_dataset_loader_form=UpdateObjectForm(new_visibility=visibility.name),
    )
    set_visibility.name()
    set_visibility.visibility()
    dataset_loader = send_graphql_request(op=op, json_response=False).update_dataset_loader
    return dataset_loader


def list_dataset_loader_version(name: str) -> pd.DataFrame:
    """
    List available versions of the dataset_loader with such name.

    Parameters
    ----------
    name: str
        Name of the DatasetLoader.

    Returns
    -------
    pd.DataFrame
        Pandas dataframe with a list of available dataset_loader versions.
    """
    op = Operation(schema.Query)
    base_query = op.dataset_loader_from_name(name=name).list_dataset_loader_version
    base_query.version()
    base_query.creation_timestamp()
    base_query.name()
    json_data = send_graphql_request(op)

    df = pd.DataFrame.from_dict(json_data["datasetLoaderFromName"]["listDatasetLoaderVersion"])
    df = _to_datetime(df, ["creationTimestamp"])

    return df.sort_values(by=["version"], ignore_index=True)


def delete_dataset_loader(dataset_loader_name: str) -> bool:
    """
    Delete dataset loader and all of it's versions.

    Parameters
    ----------
    dataset_loader_name: str
        Name of the dataset loader to delete.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    op.delete_dataset_loader(name=dataset_loader_name)
    return send_graphql_request(op)["deleteDatasetLoader"]


def delete_dataset_loader_version(dataset_loader_name: str, dataset_loader_version: int) -> bool:
    """
    Delete version of a dataset loader.

    Parameters
    ----------
    dataset_loader_name: str
        The name of the dataset loader.
    dataset_loader_version: int
        The version of the dataset loader.

    Returns
    -------
    bool
        Operation success status.
    """
    op = Operation(schema.Mutation)
    dataset_loader_version_choice = schema.ObjectVersionInput(name=dataset_loader_name, version=dataset_loader_version)
    op.delete_dataset_loader_version_from_name_version(dataset_loader_version=dataset_loader_version_choice)
    return send_graphql_request(op, json_response=False).delete_dataset_loader_version_from_name_version


def print_dataset_loader_schema(name: str, version: Optional[int] = None) -> None:
    """
    Print DatasetLoader schema.

    Parameters
    ----------
    name: str
        Name of the DatasetLoader.
    version: Optional[int] = None
        Version of the DatasetLoader. Default: None, "latest" version is used.
    """
    op = Operation(schema.Query)
    _datasetloader_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=_datasetloader_version)
    base_query.dataset_loader_method_schemas()
    json_data = send_graphql_request(op)
    json_data = json_data["datasetLoaderVersionFromNameVersion"]["datasetLoaderMethodSchemas"]
    print(f"DatasetLoader {name} version {version} json-schema:")
    for method_name, schema_ in json_data.items():
        _print_params_by_schema(json_schema=schema_, schema_type=DatasetLoaderMethodName(method_name).name)


def set_dataset_loader_version_description(name: str, version: int, description: str) -> DatasetLoaderVersionInfo:
    """
    Set dataset loader version description.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    description: str
        Description dataset loader version.

    Returns
    -------
    DatasetLoaderVersionInfo
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_description = op.update_dataset_loader_version(
        dataset_loader_version=choice, update_dataset_loader_version_form=UpdateObjectForm(new_description=description)
    )
    set_description.name()
    set_description.version()
    set_description.description()

    dataset_loader = send_graphql_request(op=op, json_response=False).update_dataset_loader_version
    return dataset_loader


def set_dataset_loader_version_visibility(
    name: str, version: int, visibility: VisibilityOptions
) -> DatasetLoaderVersionInfo:
    """
    Set dataset loader version visibility.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    visibility: VisibilityOptions
        Visibility dataset loader version.

    Returns
    -------
    DatasetLoaderVersionInfo
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_visibility = op.update_dataset_loader_version(
        dataset_loader_version=choice,
        update_dataset_loader_version_form=UpdateObjectForm(new_visibility=visibility.name),
    )
    set_visibility.name()
    set_visibility.version()
    set_visibility.visibility()

    dataset_loader = send_graphql_request(op=op, json_response=False).update_dataset_loader_version
    return dataset_loader


def set_dataset_loader_version_tags(name: str, version: int, key: str, values: list[str]) -> DatasetLoaderVersionInfo:
    """
    Set dataset loader version tags.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    key: str
        Key tag.
    values: list[str]
        Value tag.

    Returns
    -------
    DatasetLoaderVersionInfo
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_tag = op.set_dataset_loader_version_tags(dataset_loader_version=choice, key=key, values=values)
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    dataset_loader = send_graphql_request(op=op, json_response=False).set_dataset_loader_version_tags
    return dataset_loader


def reset_dataset_loader_version_tags(
    name: str, version: int, key: str, values: list[str], new_key: Optional[str] = None
) -> DatasetLoaderVersionInfo:
    """
    Reset dataset loader version tags.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    key: str
        Key tag.
    values: list[str]
        Value tag.
    new_key: Optional[str] = None
        New key of a tag.

    Returns
    -------
    DatasetLoaderVersionInfo
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    set_tag = op.reset_dataset_loader_version_tags(
        dataset_loader_version=choice, key=key, values=values, new_key=new_key
    )
    set_tag.name()
    set_tag.version()
    set_tag.tags()
    dataset_loader = send_graphql_request(op=op, json_response=False).reset_dataset_loader_version_tags
    return dataset_loader


def delete_dataset_loader_version_tag(
    name: str, version: int, key: str, value: Optional[str] = None
) -> DatasetLoaderVersionInfo:
    """
    Delete dataset loader version tag.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: int
        Version of the dataset loader.
    key: str
        Key tag.
    value: Optional[str]=None
        value tag.

    Returns
    -------
    DatasetLoaderVersionInfo
        Dataset loader version instance with meta information.
    """
    op = Operation(schema.Mutation)
    choice = schema.ObjectVersionInput(name=name, version=version)
    delete_tag = op.delete_dataset_loader_version_tag(dataset_loader_version=choice, key=key, value=value)
    delete_tag.name()
    delete_tag.version()
    delete_tag.tags()
    dataset_loader = send_graphql_request(op=op, json_response=False).delete_dataset_loader_version_tag
    return dataset_loader


def get_dataset_loader_version(name: str, version: Optional[int] = None) -> DatasetLoaderVersionInfo:
    """
    Meta information about the dataset loader version by the dataset loader name and version.

    Parameters
    ----------
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the dataset loader. Default: None, "latest" version is used.

    Returns
    -------
    DatasetLoaderVersionInfo
        DatasetLoaderVersion instance with meta information.
    """
    op = Operation(schema.Query)
    dataset_loader_version_choice = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=dataset_loader_version_choice)
    base_query.name()
    base_query.version()
    base_query.tags()
    base_query.description()
    base_query.creation_timestamp()
    data = send_graphql_request(op, json_response=False)
    return data.dataset_loader_version_from_name_version


def get_dataset_loader_version_conda_env(name: str, version: int) -> dict:
    """
    Condas configuration for the dataset loader version by the dataset loader name and version.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: Optional[int] = None
        Version of the dataset loader. Default: None, "latest" version is used.

    Returns
    -------
    Dict
        Dict with conda configuration.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=_model_version)
    base_query.get_conda_env()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.dataset_loader_version_from_name_version.get_conda_env


def get_dataset_loader_version_requirements(name: str, version: int) -> list:
    """
    Requirements for the dataset loader version by the dataset loader name and version.

    Parameters
    ----------
    name: str
        Name of the dataset loader.
    version: Optional[int] = None
        Version of the dataset loader. Default: None, "latest" version is used.

    Returns
    -------
    List
        List of requirements.
    """
    op = Operation(schema.Query)
    _model_version = schema.ObjectVersionOptionalInput(name=name, version=version)
    base_query = op.dataset_loader_version_from_name_version(dataset_loader_version=_model_version)
    base_query.list_requirements()
    model_version = send_graphql_request(op, json_response=False)
    return model_version.dataset_loader_version_from_name_version.list_requirements


def get_latest_dataset_loader_version(name: str) -> DatasetLoaderVersionInfo:
    """
    Latest dataset loader version by the dataset loader name.

    Parameters
    ----------
    name: str
        Name of the dataset loader.

    Returns
    -------
    DatasetLoaderVersionInfo
        DatasetLoaderVersion instance with meta information.
    """
    return get_dataset_loader_version(name)


def get_initial_dataset_loader_version(name: str) -> DatasetLoaderVersionInfo:
    """
    Initial dataset loader version by the dataset loader name.

    Parameters
    ----------
    name: str
        Name of the dataset loader.

    Returns
    -------
    DatasetLoaderVersionInfo
        DatasetLoaderVersion instance with meta information.
    """
    op = Operation(schema.Query)
    version = op.dataset_loader_from_name(name=name).init_dataset_loader_version()
    version.name()
    version.version()
    version.tags()
    version.description()
    dataset_loader_version = send_graphql_request(op, json_response=False)
    return dataset_loader_version.dataset_loader_from_name.init_dataset_loader_version


def pagination_dataset_loader(
    name: Optional[str] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[VisibilityOptions] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[DatasetLoaderInfo]:
    """
    Search dataset loaders.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the dataset loader.
    tag_key: Optional[str]=None
        Key of the dataset loader tag.
    tag_value: Optional[str]=None
        Value of the dataset loader tag.
    description: Optional[str]=None
        Description of the dataset loader.
    visibility: Optional[str]=None
        Visibility of dataset loader.
    owner_ids: Optional[list[str]]=None
        Ids of the dataset loader owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the dataset loader.
    creation_to: Optional[int]=None
        Creation timestamp from of the dataset loader.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the dataset loader.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the dataset loader.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection.


    Returns
    -------
    List[DatasetLoaderInfo]
        List of DatasetLoaderInfo instance with meta information.
    """
    op = Operation(schema.Query)
    base_query = op.pagination_dataset_loader(
        limit=limit,
        offset=offset,
        filter_settings=ObjectFilterSettings(
            name=name,
            description=description,
            visibility=visibility,
            owner_ids=owner_ids,
            tag=TagFilterSettings(key=tag_key, value=tag_value),
            creation_interval=TimestampInterval(start=creation_from, end=creation_to),
            last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
        ),
    ).list_dataset_loader
    base_query.name()
    base_query.tags()
    base_query.description()
    return send_graphql_request(op, json_response=False).pagination_dataset_loader.list_dataset_loader


def pagination_dataset_loader_version(
    name: str,
    version: Optional[int] = None,
    tag_key: Optional[str] = None,
    tag_value: Optional[str] = None,
    description: Optional[str] = None,
    visibility: Optional[VisibilityOptions] = None,
    owner_ids: Optional[list[str]] = None,
    creation_from: Optional[int] = None,
    creation_to: Optional[int] = None,
    last_updated_from: Optional[int] = None,
    last_updated_to: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List[DatasetLoaderVersionInfo]:
    """
    Search dataset loader versions.

    Parameters
    ----------
    name: Optional[str]=None
        Name of the dataset loader version.
    version: Optional[int] = None
        Version of the dataset loader version.
    tag_key: Optional[str]=None
        Key of the dataset loader version tag.
    tag_value: Optional[str]=None
        Value of the dataset loader version tag.
    description: Optional[str]=None
        Description of the dataset loader version.
    visibility: Optional[str]=None
        Visibility of dataset loader version.
    owner_ids: Optional[list[str]]=None
        Ids of the dataset loader version owner.
    creation_from: Optional[int]=None
        Creation timestamp from of the dataset loader version.
    creation_to: Optional[int]=None
        Creation timestamp from of the dataset loader version.
    last_updated_from: Optional[int]=None
        Last updated timestamp from of the dataset loader version.
    last_updated_to: Optional[int]=None
        Last updated timestamp from of the dataset loader version.
    limit: Optional[int] = None
        The maximum number of records that will be returned as a result.
    offset: Optional[int] = None
        The number of records that will be skipped before starting the selection

    Returns
    -------
    List[DatasetLoaderVersionInfo]
        List of DatasetLoaderVersionInfo instance with meta information.
    """
    op = Operation(schema.Query)
    base_query = (
        op.dataset_loader_from_name(name=name)
        .pagination_dataset_loader_version(
            limit=limit,
            offset=offset,
            filter_settings=ObjectVersionFilterSettings(
                version=version,
                description=description,
                visibility=visibility,
                owner_ids=owner_ids,
                tag=TagFilterSettings(key=tag_key, value=tag_value),
                creation_interval=TimestampInterval(start=creation_from, end=creation_to),
                last_updated_interval=TimestampInterval(start=last_updated_from, end=last_updated_to),
            ),
        )
        .list_dataset_loader_version
    )
    base_query.name()
    base_query.version()
    base_query.tags()
    base_query.description()
    return send_graphql_request(
        op, json_response=False
    ).dataset_loader_from_name.pagination_dataset_loader_version.list_dataset_loader_version



"""Module for automatic jsonschema inference."""
import typing
import warnings
from copy import copy
from inspect import formatannotation, getfullargspec

from jsonschema import SchemaError
from jsonschema.validators import Draft4Validator

from ML_management.mlmanagement.jsonschema_exceptions import (
    DictKeysMustBeStringsError,
    FunctionContainsVarArgsError,
    FunctionContainsVarKwArgsError,
    InvalidSchemaError,
    InvalidStructureAnnotationError,
    NoAnnotationError,
    UnsupportedTypeError,
)

JSON_SCHEMA_DRAFT = "http://json-schema.org/draft-04/schema#"

type_map = {
    "bool": {"type": "boolean"},
    "int": {"type": "integer"},
    "float": {"type": "number"},
    "str": {"type": "string"},
}


class SkipJsonSchema:
    """
    Annotation wrapper for system parameters to skip them in JSON schema.

    Usage example::

        def my_function(arg: Annotated[np.ndarray, SkipJsonSchema]) -> None:
            print(arg)

    """

    def __class_getitem__(cls, _type):  # for backward compatibility with older models
        return cls()


def __get_or_raise(type_name, from_optional: bool = False):
    if type_name in type_map:
        field = copy(type_map[type_name])
        if from_optional:
            field["type"] = [field["type"], "null"]
        return field
    else:
        raise UnsupportedTypeError(annotation=type_name, supported_types=list(type_map.keys()))


def __is_optional(annotation):
    repr_annotation = repr(annotation)
    if repr_annotation.startswith("typing.Optional"):
        return True
    elif repr_annotation.startswith("typing.Union") and repr_annotation.strip("]").endswith("NoneType"):
        # python 3.8 casts Optional to Union[arg, None]
        return True
    else:
        return False


def __get_json_schema_from_annotation(annotation, from_optional: bool = False):
    if annotation.__module__ == "typing":
        if __is_optional(annotation):
            return __get_json_schema_from_annotation(
                annotation.__args__[0], from_optional=True
            )  # Optional[int] translates to type: [integer, null], and the field is not required
        else:
            formatted_annotation = formatannotation(annotation)

            # this is the only way to reliably get annotation name (e.g. "List")
            annotation_name = formatted_annotation.partition("[")[0]
            if not hasattr(annotation, "__args__"):
                raise InvalidStructureAnnotationError(annotation=formatted_annotation)
            if annotation_name == "List":
                return {
                    "type": "array" if not from_optional else ["array", "null"],
                    "items": __get_json_schema_from_annotation(annotation.__args__[0]),
                }
            elif annotation_name == "Dict":
                key_annotation = annotation.__args__[0]
                if hasattr(key_annotation, "__name__"):
                    if key_annotation.__name__ != "str":
                        raise DictKeysMustBeStringsError(annotation=formatted_annotation)
                else:
                    raise DictKeysMustBeStringsError(annotation=formatted_annotation)
                return {
                    "type": "object" if not from_optional else ["object", "null"],
                    "additionalProperties": __get_json_schema_from_annotation(annotation.__args__[1]),
                }
            elif annotation_name == "Union":
                union_types = [__get_json_schema_from_annotation(ann) for ann in annotation.__args__]
                marked_as_null = False
                for type_ in union_types:
                    if type_ == {"type": "null"}:
                        marked_as_null = True
                        break
                if not marked_as_null and from_optional:
                    union_types.extend([{"type": "null"}])
                return {"anyOf": union_types}
            else:
                raise UnsupportedTypeError(
                    annotation=formatannotation(annotation),
                    supported_types=list(type_map.keys()),
                )

    elif annotation.__module__ == "builtins":
        if hasattr(annotation, "__name__"):
            return __get_or_raise(annotation.__name__, from_optional=from_optional)
    else:
        raise UnsupportedTypeError(
            annotation=formatannotation(annotation),
            supported_types=list(type_map.keys()),
        )


def get_function_args(spec, func):
    """Get all arguments of function."""
    # FullArgSpec(args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations)
    pos_args_defaults = {}
    if spec.defaults:
        for arg, default in zip(spec.args[-len(spec.defaults) :], spec.defaults):
            pos_args_defaults[arg] = default
    if spec.varargs:
        raise FunctionContainsVarArgsError(function_name=func.__name__)
    if spec.varkw:
        raise FunctionContainsVarKwArgsError(function_name=func.__name__)

    # annotations are not inferred from default yet, but i can do it myself from kwonlydefaults
    pos_args = spec.args if spec.args else []
    kw_only_args = spec.kwonlyargs if spec.kwonlyargs else []
    all_args = pos_args + kw_only_args
    kw_only_args_defaults = spec.kwonlydefaults if spec.kwonlydefaults else {}
    all_defaults = {**pos_args_defaults, **kw_only_args_defaults}

    return all_args, all_defaults


def infer_jsonschema(func, get_object_func):
    """Infer jsonschema from callable by its signature."""
    from ML_management.model.model_type_to_methods_map import ModelMethodName  # circular import

    spec = getfullargspec(func)
    all_args, all_defaults = get_function_args(spec, func)
    schema = {
        "$schema": JSON_SCHEMA_DRAFT,
        "type": "object",
        "properties": {},
        "additionalProperties": False,
    }
    if "self" in all_args:
        all_args.remove("self")

    if func.__name__ == ModelMethodName.init.value and get_object_func:
        spec_get_object = getfullargspec(get_object_func)
        all_args_get_object, _ = get_function_args(spec_get_object, get_object_func)
        if not set(all_args_get_object) == set(all_args):
            warnings.warn(
                "get_object function arguments do not match __init__ arguments. "
                "You won't be able to pass none default arguments into __init__ function."
            )
            all_args = []
            all_defaults = []

    required = []
    for arg in all_args:
        if arg not in spec.annotations:
            raise NoAnnotationError(arg=arg)

        # skip system parameters
        if (
            typing.get_origin(spec.annotations[arg]) is typing.Annotated
            and spec.annotations[arg].__metadata__
            and spec.annotations[arg].__metadata__[0] is SkipJsonSchema
        ) or isinstance(spec.annotations[arg], SkipJsonSchema):
            continue

        schema["properties"][arg] = __get_json_schema_from_annotation(spec.annotations[arg])
        if arg in all_defaults:
            schema["properties"][arg]["default"] = all_defaults[arg]
        if arg not in all_defaults and not __is_optional(spec.annotations[arg]):
            required.append(arg)

    if required:
        schema["required"] = required
    try:
        Draft4Validator.check_schema(schema)
    except SchemaError as err:
        raise InvalidSchemaError(schema=schema, original_message=str(err)) from None

    return schema



"""Module with pydantics model representing parameters of sdk functions."""

from typing import List, Literal, Optional

from pydantic import BaseModel, PositiveInt

from ML_management.dataset_loader.dataset_loader_pattern_to_methods_map import DatasetLoaderMethodName
from ML_management.executor import BaseExecutor
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.model.model_type_to_methods_map import ModelMethodName


class ModelMethodParams(BaseModel):
    """Class for parameters for one model method.

    Attributes
    ----------
    method: ModelMethodName
        method of model.
    params: dict
        parameters for method.
        Example::

            {key1: value1, key2: value2}

    """

    method: ModelMethodName
    params: dict

    def serialize(self) -> dict:
        return {self.method.value: self.params}


class DatasetLoaderMethodParams(BaseModel):
    """Class for parameters for one dataset loader method.

    Attributes
    ----------
    method: DatasetLoaderMethodName
        method of dataset loader.
    params: dict
        parameters for method.
        Example::

            {key1: value1, key2: value2}

    """

    method: DatasetLoaderMethodName = DatasetLoaderMethodName.get_dataset
    params: dict

    def serialize(self) -> dict:
        return {self.method.value: self.params}


class ModelVersionChoice(BaseModel):
    """Class for model version choice in add_ml_job.

    Attributes
    ----------
    name: str
        Name of the model to interact with.
    version: Optional[int] = None
        Version of the model to interact with.
        Default: None, "latest" version is used.
    choice_criteria: {"latest", "initial"}
        Criteria to choose the model to interact with. Required if model_version is not specified. Default: "latest"
        If "best": according to the optimal_min value, minimum (or maximum) value for the selected metric_name is taken.
    """

    name: str
    version: Optional[PositiveInt] = None
    choice_criteria: Literal["latest", "initial"] = "latest"

    def serialize(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "choice_criteria": self.choice_criteria,
        }


class ModelForm(BaseModel):
    """Class for single model parameters for add_ml_job function.

    Attributes
    ----------
    model_version_choice: ModelVersionChoice
        parameters that determine which model to use.
    params: List[ModelMethodParams] = []
        List of ModelMethodParams with parameters of model methods.
        Example::

            [
                ModelMethodParams(method=ModelMethodName.evaluate_function, params={key1: value1, key2: value2}),
                ModelMethodParams(method=ModelMethodName.finetune_function, params={key1: value1, key2: value2})
            ]

        Default: [].
    new_model_name: Optional[str] = None
        Name of the model to save in case new model is to be created as a result of job execution.
        (regulated by executor.upload_model_mode).
        Default: None.
    new_model_description: Optional[str] = None
        Description of the new model name.
        Default: None.
    prepare_new_model_inference: bool = False
        Start preparing the environment for the inference of the new model.
        Default: False.
    start_build_new_model_image: bool = True
        Start build image for the new model.
        Default: True.
    new_model_visibility: VisibilityOptions
        Visibility of the new model to other users. Default: PRIVATE.
    """

    model_version_choice: ModelVersionChoice
    params: List[ModelMethodParams] = []
    new_model_name: Optional[str] = None
    new_model_description: Optional[str] = None
    prepare_new_model_inference: bool = False
    start_build_new_model_image: bool = True
    new_model_visibility: VisibilityOptions = VisibilityOptions.PRIVATE

    def serialize(self) -> List[dict]:
        return [
            {
                "role": BaseExecutor.DEFAULT_ROLE,
                "model": self.model_version_choice.serialize(),
                "params": [method_params.serialize() for method_params in self.params],
                "new_model_name": self.new_model_name,
                "prepare_new_model_inference": self.prepare_new_model_inference,
                "start_build_new_model_image": self.start_build_new_model_image,
                "new_model_visibility": self.new_model_visibility,
                "description": self.new_model_description,
            }
        ]


class ModelWithRole(ModelForm):
    """Class for one role model with parameters for add_ml_job function.

    Same as ModelForm except parameter role: str is required.

    Attributes
    ----------
    model_version_choice: ModelVersionChoice
        parameters that determine which model to use.
    params: List[ModelMethodParams] = []
        List of ModelMethodParams with parameters of model methods.
        Example::

            [
                ModelMethodParams(method=ModelMethodName.evaluate_function, params={key1: value1, key2: value2}),
                ModelMethodParams(method=ModelMethodName.finetune_function, params={key1: value1, key2: value2})
            ]

        Default: [].
    new_model_name: Optional[str] = None
        Name of the model to save in case new model is to be created as a result of job execution.
        (regulated by executor.upload_model_mode).
        Default: None.
    new_model_description: Optional[str] = None
        Description of the new model name.
        Default: None.
    prepare_new_model_inference: bool = False
        Start preparing the environment for the inference of the new model.
        Default: False.
    role: str
        Role of the model.
    """

    role: str

    def serialize(self) -> List[dict]:
        result = super().serialize()
        result[0]["role"] = self.role
        return result


class AnyModelForm(BaseModel):
    """Class for definition of arbitrary number of models used in job with their roles.

    Attributes
    ----------
    models: List[ModelWithRole]

        Example (Both instances of ModelWithRole and usual list of dicts may be used)::

            [
                ModelWithRole(
                    role="role_one",
                    model_version_choice=ModelVersionChoice(
                        name="model_1_name",
                        version=1,  # Optional. Default: latest version
                    )
                    params=[
                        ModelMethodParams(
                            method=ModelMethodName.evaluate_function,
                            params={key1: value1, key2: value2}
                        ),
                        ModelMethodParams(
                            method=ModelMethodName.finetune_function,
                            params={key1: value1, key2: value2}
                        )
                    ]
                    new_model_name="some_name"  # Optional. Default: None
                    new_model_description="some_description" # Optional. Default: None
                ),
                ModelWithRole(
                    role="role_two",
                    model_version_choice=ModelVersionChoice(name="model_2_name"),
                    params=[
                        ModelMethodParams(
                            method=ModelMethodName.evaluate_function,
                            params={key1: value1, key2: value2}
                        ),
                        ModelMethodParams(
                            method=ModelMethodName.finetune_function,
                            params={key1: value1, key2: value2}
                        )
                    ]
                )
            ]

    """

    models: List[ModelWithRole]

    def serialize(self) -> List[dict]:
        if len(self.models) == 0:
            return self.models

        result = self.models[0].serialize()
        for model in self.models[1:]:
            result.extend(model.serialize())
        return result


class DatasetLoaderForm(BaseModel):
    """Class for single dataset loader parameters for add_ml_job function.

    Attributes
    ----------
    name: str
        Name of the DatasetLoader that the model will use.
    version: Optional[int] = None
        Version of the DatasetLoader that the model will interact with.
        Default: None, "latest" version is used.
    params: List[DatasetLoaderMethodParams] = []
        List of DatasetLoaderMethodParams with parameters of model methods.
        Example::

            [
                DatasetLoaderMethodParams(
                    # by default params are set for DatasetLoaderMethodName.get_dataset method
                    params={key1: value1, key2: value2}
                ),
                DatasetLoaderMethodParams(
                    method=DatasetLoaderMethodName.<some_method>,
                    params={key1: value1, key2: value2}
                )
            ]

        Default: [].
    collector_name: {"s3", }
        Name of the collector to interact with. Default: "s3"
    collector_params: dict
        Dictionary of collector parameters.
        Example::

            {"bucket": "mnist"}

    """

    name: str
    params: List[DatasetLoaderMethodParams] = []
    collector_params: dict
    version: Optional[PositiveInt] = None
    collector_name: str = "s3"

    def serialize(self) -> List[dict]:
        return [
            {
                "role": BaseExecutor.DEFAULT_ROLE,
                "data_params": {
                    "collector_name": self.collector_name,
                    "collector_params": self.collector_params,
                    "dataset_loader_name": self.name,
                    "dataset_loader_version": self.version,
                    "dataset_loader_params": [method_params.serialize() for method_params in self.params],
                },
            }
        ]


class DatasetLoaderWithRole(DatasetLoaderForm):
    """Class for one role dataset loader with parameters for add_ml_job function.

    Same as DatasetLoaderForm except parameter role: str is required.

    Attributes
    ----------
    name: str
        Name of the DatasetLoader that the model will use.
    version: Optional[int] = None
        Version of the DatasetLoader that the model will interact with.
        Default: None, "latest" version is used.
    params: List[DatasetLoaderMethodParams] = []
        List of DatasetLoaderMethodParams with parameters of model methods.
        Example::

            [
                DatasetLoaderMethodParams(
                    # by default params are set for DatasetLoaderMethodName.get_dataset method
                    params={key1: value1, key2: value2}
                ),
                DatasetLoaderMethodParams(
                    method=DatasetLoaderMethodName.<some_method>,
                    params={key1: value1, key2: value2}
                )
            ]

        Default: [].
    collector_name: {"s3", }
        Name of the collector to interact with.
        Default: "s3"
    collector_params: dict
        Dictionary of collector parameters.
        Example::

            {"bucket": "mnist"}

    role: str
        Role of the DatasetLoader.
    """

    role: str

    def serialize(self) -> List[dict]:
        result = super().serialize()
        result[0]["role"] = self.role
        return result


class AnyDatasetLoaderForm(BaseModel):
    """Class for definition of arbitrary number of dataset_loaders used in job with their roles.

    Attributes
    ----------
    datasetloaders: List[DatasetLoaderWithRole]

        Example (Both instances of DatasetLoaderWithRole and usual list of dicts may be used)::

            [
                DatasetLoaderWithRole(
                    role="data1",
                    name="multiple_two_datasets",
                    version=1,  # Optional. Default: latest version
                    params=[
                        DatasetLoaderMethodParams(
                            # by default params are set for DatasetLoaderMethodName.get_dataset method
                            params={key1: value1, key2: value2}
                        ),
                        DatasetLoaderMethodParams(
                            method=DatasetLoaderMethodName.get_dataset,  # or other method
                            params={key1: value1, key2: value2}
                        )
                    ],
                    collector_name="s3",
                    collector_params={"bucket": "mnist"}
                ),
                DatasetLoaderWithRole(
                    role="data2",
                    name="multiple_two_datasets",
                    collector_name="s3",
                    collector_params={"bucket": "mnist"}
                )
            ]

    """

    dataset_loaders: List[DatasetLoaderWithRole]

    def serialize(self) -> List[dict]:
        if len(self.dataset_loaders) == 0:
            return []

        result = self.dataset_loaders[0].serialize()
        for model in self.dataset_loaders[1:]:
            result.extend(model.serialize())
        return result



import time
import warnings
from collections import deque
from threading import Event, Thread
from typing import Dict

from sgqlc.operation import Operation

from ML_management import variables
from ML_management.graphql import schema
from ML_management.graphql.schema import MetricInput
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.singleton_pattern import Singleton
from ML_management.variables import (
    get_butch_polling_frequency,
    get_metric_accumulation_duration,
    get_timeout_log_metric_batch,
)


class Batcher(metaclass=Singleton):
    def __init__(self):
        self.batch = deque()
        self.thread = None
        self.stop_event = Event()

    def log_metric(self, key: str, value: float, step: int, timestamp: int):
        self.batch.append((key, value, step, timestamp))
        self._start_thread()

    def log_metrics(self, metrics: Dict[str, float], step: int, timestamp: int):
        self.batch.extend((key, value, step, timestamp) for key, value in metrics.items())
        self._start_thread()

    def _start_thread(self):
        if not self.thread:
            self.stop_event.clear()
            self.thread = Thread(target=self.log_batch_metric, args=(self.stop_event,), daemon=True)
            self.thread.start()

    def _stop_batching(self):
        self.batch = deque()
        if self.thread:
            self.stop_event.set()
            self.stop_event = Event()
            self.thread = None

    def log_batch_metric(self, stop_event):
        metric_accumulation_duration = get_metric_accumulation_duration()
        while not stop_event.is_set():
            time.sleep(metric_accumulation_duration)
            if len(self.batch) == 0:
                continue
            metrics_to_log = self.batch.copy()
            op = Operation(schema.Mutation)
            op.log_metrics(
                metrics=[
                    MetricInput(key=key, value=value, step=step, timestamp=timestamp)
                    for key, value, step, timestamp in metrics_to_log
                ],
                secret_uuid=variables.secret_uuid,
            )
            try:
                send_graphql_request(op, json_response=False)
                for _ in range(len(metrics_to_log)):
                    self.batch.popleft()
            except Exception as err:
                warnings.warn(str(err))

    def wait_log_metrics(self):
        stop_time = time.time()
        butch_polling_frequency = get_butch_polling_frequency()
        while self.batch and time.time() - stop_time < get_timeout_log_metric_batch():
            time.sleep(butch_polling_frequency)
        self._stop_batching()



"""Class inherits from base class of Model with a specific method for obtaining gradients."""
from abc import ABC, abstractmethod
from typing import Annotated, Callable

import numpy as np

from ML_management.mlmanagement.jsonschema_inference import SkipJsonSchema
from ML_management.model.patterns import model_pattern


class GradientModel(model_pattern.Model, ABC):
    """Implementation of gradient model."""

    @abstractmethod
    def get_grad(
        self, loss_fn: Annotated[Callable, SkipJsonSchema], input_batch: Annotated[np.ndarray, SkipJsonSchema]
    ) -> np.ndarray:
        """Define get_grad function."""
        raise NotImplementedError



"""Abstract base class for model transformation methods."""
from abc import ABC, abstractmethod
from typing import Annotated, Callable

from ML_management.mlmanagement.jsonschema_inference import SkipJsonSchema
from ML_management.model.patterns.model_pattern import Model


class Transformer(Model, ABC):
    """Abstract class for model that performs transformations."""

    @abstractmethod
    def transform(self, model_fn: Annotated[Callable, SkipJsonSchema], **kwargs) -> Callable:
        """Perform model transformation.

        :param model_fn: takes a batch of input tensors and produces a final prediction tensor.
        """
        raise NotImplementedError



import importlib
import os
import os.path
import subprocess
import sys
import tarfile
import tempfile
import threading
import traceback
from pathlib import Path
from typing import Optional

import yaml

from ML_management.base_exceptions import MLMClientError
from ML_management.mlmanagement.log_api import _raise_error
from ML_management.mlmanagement.model_type import ModelType
from ML_management.session import AuthSession
from ML_management.variables import (
    CONFIG_KEY_ARTIFACTS,
    DATA,
    FLAVOR_NAME,
    LEGACY_DATA,
    MLCONFIG,
    MLMODEL_FILE_NAME,
    get_log_service_url,
)


def download_artifacts_by_name_version(
    name: str,
    version: Optional[int],
    model_type: ModelType,
    path: str,
    dst_path: Optional[str] = None,
) -> str:
    """Download an artifact by name and version to a local directory, and return a local path for it.

    Parameters
    ==========
    name: str
        Name of the entity.
    version: Optional[int] = None
        Version of the entity. Default: None, "latest" version is used.
    model_type: ModelType
        Type of the entity. Possible values: ModelType.MODEL | ModelType.EXECUTOR | ModelType.DATASET_LOADER
    path: str = ""
        Specific path for artifacts download. Default: "", all artifacts will be downloaded.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    Returns
    =======
    str
        Local path to the entity folder.
    """
    url = get_log_service_url("download_artifacts_by_name_version")
    params = {
        "path": os.path.normpath(path) if path else path,
        "name": name,
        "model_type": model_type.value,
    }
    if version:
        params["version"] = version
    return _request_download_artifacts(url, params, dst_path)


def download_job_artifacts(name: str, path: str = "", dst_path: Optional[str] = None) -> str:
    """Download an artifact file or directory from a job to a local directory, and return a local path for it.

    Parameters
    ==========
    name: str
        Name of the job.
    path: str = ""
        Specific path for artifacts download. Default: "", all artifacts will be downloaded.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    Returns
    =======
    str
        Local path to artifacts.
    """
    url = get_log_service_url("download_job_artifacts")
    params = {"path": os.path.normpath(path) if path else path, "job_name": name}
    return _request_download_artifacts(url, params, dst_path)


def _load_model_type(
    name: str,
    version: Optional[int],
    model_type: ModelType,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init=None,
):
    """Load model from local path."""
    local_path = download_artifacts_by_name_version(
        name=name, version=version, model_type=model_type, path="", dst_path=dst_path
    )
    if install_requirements:
        _set_model_version_requirements(local_path)
    loaded_model = _load_model_src(local_path, kwargs_for_init)
    return loaded_model


def load_dataset(
    name: str,
    version: Optional[int] = None,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init: Optional[dict] = None,
):
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the dataset.
    version: Optional[int] = None
        Version of the dataset. Default: None, "latest" version is used.
    install_requirements: bool = False
        Whether to install dataset requirements. Default: False.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    kwargs_for_init: Optional[dict]: None
        Kwargs for __init__ function of dataset loader when it would be loaded.
    Returns
    =======
    DatasetLoaderPattern
        The object of the dataset to use.
    """
    return _load_model_type(
        name,
        version,
        ModelType.DATASET_LOADER,
        install_requirements,
        dst_path,
        kwargs_for_init,
    )


def _set_model_version_requirements(local_path) -> None:
    """Installing requirements of the model locally."""
    with open(os.path.join(local_path, "requirements.txt")) as req:
        requirements = list(
            filter(
                lambda x: "ml-management" not in x.lower() and "mlflow" not in x.lower() and len(x),
                req.read().split("\n"),
            )
        )
    try:
        if requirements:
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "--no-cache-dir",
                    "--default-timeout=100",
                    *requirements,
                ]
            )

    except Exception:
        print(traceback.format_exc())


def load_model(
    name: str,
    version: Optional[int] = None,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
    kwargs_for_init=None,
):
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the model.
    version: Optional[int] = None
        Version of the model. Default: None, "latest" version is used.
    install_requirements: bool = False
        Whether to install model requirements. Default: False.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    kwargs_for_init: Optional[dict]: None
        Kwargs for __init__ function of model when it would be loaded.
    Returns
    =======
    Model
        The object of the model to use.
    """
    return _load_model_type(name, version, ModelType.MODEL, install_requirements, dst_path, kwargs_for_init)


def load_executor(
    name: str,
    version: Optional[int] = None,
    install_requirements: bool = False,
    dst_path: Optional[str] = None,
):
    """Download all model's files for loading model locally.

    Parameters
    ==========
    name: str
        Name of the executor.
    version: Optional[int] = None
        Version of the executor. Default: None, "latest" version is used.
    install_requirements: bool = False
        Whether to install executor requirements. Default: False.
    dst_path: Optional[str]: None
        Destination path. Default: None.
    Returns
    =======
    BaseExecutor
        The object of the executor to use.
    """
    return _load_model_type(name, version, ModelType.EXECUTOR, install_requirements, dst_path)


def _untar_folder(buff, to_folder):
    try:
        with tarfile.open(mode="r|", fileobj=buff) as tar:
            tar.extractall(to_folder)
    except Exception as err:
        raise MLMClientError("Some error during untar the content.") from err


def _request_download_artifacts(url, params: dict, dst_path: Optional[str] = None):
    path = params["path"]
    with AuthSession().get(url=url, params=params, stream=True) as response:
        _raise_error(response)
        untar = response.headers.get("untar") == "True"
        if dst_path is None:
            dst_path = tempfile.mkdtemp()
        dst_path = os.path.abspath(os.path.normpath(dst_path))
        local_path = os.path.normpath(os.path.join(dst_path, os.path.normpath(path)))
        if untar:
            r, w = os.pipe()
            with open(r, "rb") as buff:
                try:
                    thread = threading.Thread(target=_untar_folder, args=(buff, local_path))
                    thread.start()
                except Exception as err:
                    os.close(r)
                    os.close(w)
                    raise err

                with open(w, "wb") as wfd:
                    for chunk in response.iter_raw():
                        wfd.write(chunk)
                thread.join()
                return local_path
        else:
            dirs = os.path.dirname(local_path)
            if not os.path.exists(dirs):
                os.makedirs(dirs)
            with open(local_path, "wb") as f:
                for chunk in response.iter_bytes():
                    f.write(chunk)
            return local_path


def _load_model_src(local_path: str, kwargs_for_init: Optional[dict]):
    if not kwargs_for_init:
        kwargs_for_init = {}
    config_path = os.path.join(local_path, MLCONFIG)
    if os.path.exists(config_path):
        with open(config_path) as file:
            conf = yaml.safe_load(file)

        load_model_path = os.path.join(local_path, conf["load_model_path"])

    elif os.path.exists(os.path.join(local_path, MLMODEL_FILE_NAME)):
        with open(os.path.join(local_path, MLMODEL_FILE_NAME)) as file:
            conf = yaml.safe_load(file)

        # legacy data entity code path for backward compatibility
        code_data_path = (
            conf["flavors"][FLAVOR_NAME][DATA]
            if DATA in conf["flavors"][FLAVOR_NAME]
            else conf["flavors"][FLAVOR_NAME][LEGACY_DATA]
        )
        load_model_path = os.path.join(local_path, code_data_path)

    else:
        raise RuntimeError("MLConfig does not exist.")

    from ML_management.mlmanagement.utils import INIT_FUNCTION_NAME  # circular import

    parts = Path(load_model_path).parts
    if str(Path(*parts[:2])) not in sys.path:
        sys.path.append(str(Path(*parts[:2])))
    python_model = getattr(importlib.import_module(".".join(parts[2:])), INIT_FUNCTION_NAME)(**kwargs_for_init)
    artifacts = Path(load_model_path) / CONFIG_KEY_ARTIFACTS
    if not artifacts.exists():
        artifacts.mkdir()

    python_model.artifacts = str(artifacts)

    return python_model



"""Define simple dataset_loader class."""
from ML_management.dataset_loader.dataset_loader_pattern import DatasetLoaderPattern


class DummyDataWrapper(DatasetLoaderPattern):
    """Dummy dataset loader class."""

    def get_dataset(self):
        """Return None."""
        return None



from sgqlc.operation import Operation

from ML_management.graphql.schema import schema
from ML_management.graphql.send_graphql_request import send_graphql_request


def list_metric_jobs(metric_name: str, job_names: list[str]) -> list[dict]:
    """
    Return metrics by job names and metric_name.

    Parameters
    ----------
    metric_name: str
        Name of the metric.
    job_names: list[str]
        List of job names.

    Returns
    -------
    list[dict]
        List of metrics.
    """
    op = Operation(schema.Query)

    op.list_metric_jobs(metric_name=metric_name, job_names=job_names)

    metrics = send_graphql_request(op, json_response=False)

    return metrics.list_metric_jobs



from ML_management.mlmanagement import active_job, backend_api, load_api, log_api

log_executor_src = log_api.log_executor_src
log_dataset_loader_src = log_api.log_dataset_loader_src
log_model_src = log_api.log_model_src
log_metric = log_api.log_metric
log_metrics = log_api.log_metrics
log_artifact = log_api.log_artifact
set_server_url = backend_api.set_server_url
get_server_url = backend_api.get_server_url
set_mlm_credentials = backend_api.set_mlm_credentials
load_model = load_api.load_model
load_dataset = load_api.load_dataset
load_executor = load_api.load_executor
download_artifacts_by_name_version = load_api.download_artifacts_by_name_version
download_job_artifacts = load_api.download_job_artifacts
start_job = active_job.start_job



import traceback
from typing import Optional

from sgqlc.operation import Operation

from ML_management import variables
from ML_management.graphql import schema
from ML_management.graphql.send_graphql_request import send_graphql_request
from ML_management.mlmanagement.batcher import Batcher
from ML_management.mlmanagement.visibility_options import VisibilityOptions
from ML_management.variables import DEFAULT_EXPERIMENT


class ActiveJob:
    """
    A context manager that allows for the execution of a task locally.

    This class provides a convenient way to run a job locally.

    """

    def __init__(self, secret_uuid):
        self.secret_uuid = secret_uuid

    def __enter__(self):
        op = Operation(schema.Mutation)
        base_query = op.start_job(secret_uuid=self.secret_uuid)
        base_query.name()
        base_query.experiment.name()
        _query_job_params(base_query)
        self.job = send_graphql_request(op=op, json_response=False).start_job
        variables.secret_uuid = self.secret_uuid
        return self.job

    def __exit__(self, exc_type, exc_val, exc_tb):
        Batcher().wait_log_metrics()
        exception_traceback = None
        message = None
        status = "SUCCESSFUL"
        if exc_type:
            exception_traceback = traceback.format_exc()
            message = ": ".join([exc_type.__name__, str(exc_val)])
            status = "FAILED"

        op = Operation(schema.Mutation)
        op.stop_job(
            secret_uuid=variables.secret_uuid, status=status, message=message, exception_traceback=exception_traceback
        )
        try:
            _ = send_graphql_request(op=op, json_response=False).stop_job
        finally:
            variables.secret_uuid = None


def start_job(
    job_name: Optional[str] = None, experiment_name: str = DEFAULT_EXPERIMENT, visibility=VisibilityOptions.PRIVATE
) -> ActiveJob:
    """
    Create local job.

    Parameters
    ----------
    job_name: str | None=None
        Name of the new job. If not passed, it will be generated.
    experiment_name: str = "Default"
        Name of the experiment. Default: "Default"
    visibility: VisibilityOptions
        Visibility of this job to other users. Default: PRIVATE.

    Returns
    -------
    ActiveJob
        Active job.

    Usage:
        with start_local_job('my-beautiful-job') as job:
            mlmanagement.log_metric(...)
            mlmanagement.log_artifacts(...)
    """
    op = Operation(schema.Mutation)
    op.create_local_job(job_name=job_name, experiment_name=experiment_name, visibility=visibility.name)
    secret_uuid = send_graphql_request(op=op, json_response=False).create_local_job
    return ActiveJob(secret_uuid)


def _query_job_params(base_query):
    base_query.params()
    base_query.params.gpu()
    base_query.params.models_schemas()
    base_query.params.list_role_model_params()
    base_query.params.list_role_data_params()
    base_query.params.list_role_data_params.data_params()
    base_query.params.list_role_data_params.role()
    base_query.params.list_role_model_params.model_params()
    base_query.params.list_role_model_params.role()
    base_query.params.executor_params()
    base_query.params.executor_params.executor_method_params()
    base_query.params.executor_params.executor_version_choice()
    base_query.executor_version.upload_model_modes()



"""Define Abstract class for Model with necessary methods and methods to implement."""
import inspect
import os
from abc import ABC, abstractmethod
from pathlib import Path

from ML_management.variables import CONFIG_KEY_ARTIFACTS


class Model(ABC):
    """
    Abstract class for model that Job will use.

    Attributes
    ----------
        self.artifacts: str
            Local path to artifacts.

        self.dataset: DatasetLoader object
            Instance of user's dataset.

    """

    def __new__(cls, *args, **kwargs):  # noqa: ARG003
        """Get object of Model class."""
        self = super().__new__(cls)
        # For now it was impossible to set self.artifacts before init func of model,
        # so it was impossible to use it inside init.
        # Because inside the job container code we have a fixed folder and file structure,
        # we can predetermine artifacts path.
        # It has to be beside file __init__.py  with get_object function.
        self.artifacts = str(
            Path(os.path.dirname(inspect.getframeinfo(inspect.currentframe().f_back)[0])) / CONFIG_KEY_ARTIFACTS
        )
        self.dataset = None

        return self

    @abstractmethod
    def predict_function(self, **kwargs):
        """Every model should make predictions."""
        raise NotImplementedError

    def to_device(self, device: str) -> None:  # noqa: B027
        """
        Define model migration to specific device.

        Devices are marked with following notation:

        cpu - CPU instance

        cuda:<number: int> - GPU instance
        """
        pass



"""Class inherits from base class of Model with a specific method for getting loss from the model."""
from abc import ABC, abstractmethod

from ML_management.model.patterns import model_pattern


class ModelWithLosses(model_pattern.Model, ABC):
    """Implementation of model with loss function."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_losses(self, **kwargs):
        """Define get_losses function."""
        raise NotImplementedError
